from os import path

with open(path.join(path.dirname(__file__), "VERSION"), encoding="utf-8") as f:
    __version__ = f.read().strip()

__author__ = "Skytek Ltd."
__author_website__ = "https://skytek.com"



from typing import Optional

import requests
from shapely.geometry import Polygon

FROM_SETTINGS = object()
WGS84 = 4326


class ArcGisClient:
    def __init__(
        self,
        base_layer_url: str,
        auth_key=FROM_SETTINGS,
        add_dummy_where=True,
        timeout=300,
    ) -> None:
        if auth_key is FROM_SETTINGS:
            try:
                from django.conf import (  # pylint: disable=import-outside-toplevel
                    settings,
                )

                auth_key = settings.ARCGIS_AUTH_KEY
            except (ImportError, AttributeError):
                auth_key = None
        self.auth_key = auth_key

        self.base_layer_url = base_layer_url
        self.add_dummy_where = add_dummy_where
        self.format = "geojson"
        self.fields = ("*",)
        self.timeout = timeout

    def get_feature_list(self, bounding_polygon: Optional[Polygon] = None, params=None):
        if params is None:
            params = {}
        url = self.base_layer_url + "/query"

        if self.add_dummy_where and "where" not in params:
            params = {**params, "where": "1=1"}

        if self.fields:
            params = {**params, "outFields": ",".join(self.fields)}

        if bounding_polygon:
            params = self._geometry_params(bounding_polygon, params)

        response = self._do_request(url, params)
        if "error" in response:
            print(response)
        return response.get("features", [])

    def get_feature(self, object_id: str):
        url = self.base_layer_url + "/" + object_id
        return self._do_request(url)

    def get_info(self):
        url = self.base_layer_url
        return self._do_request(url, format="json")

    def _geometry_params(self, bounding_polygon: Polygon, params=None):
        params = params or {}
        params = {
            **params,
            "geometry": ",".join(map(str, bounding_polygon.bounds)),
            "geometryType": "esriGeometryEnvelope",
            "inSR": WGS84,
            "spatialRel": "esriSpatialRelIntersects",
        }
        return params

    def _extend_params(
        self, params=None, format=None
    ):  # pylint: disable=redefined-builtin
        params = params or {}
        params = {**params, "f": format or self.format}
        if self.auth_key:
            params = {**params, "token": self.auth_key}
        return params

    def _do_request(
        self, url, params=None, format=None
    ):  # pylint: disable=redefined-builtin
        params = self._extend_params(params, format)
        response = requests.get(url, params=params, timeout=self.timeout)
        response.raise_for_status()
        return response.json()



from base64 import b64encode

from generic_map_api.serializers import BaseFeatureSerializer


class ParametrizedArcGisSerializer(BaseFeatureSerializer):
    feature_type = "arc-gis-adhoc"

    def get_id(self, obj):
        feature_id = obj.get("id", None)
        if not feature_id:
            return None
        return f"{b64encode(obj['_layer_url'])},{feature_id}"

    def get_geometry(self, obj):
        return obj["geometry"]

    def serialize_details(self, obj):
        return {
            **obj["feature"]["attributes"],
        }


class AdHocBaseArcGisSerializer(BaseFeatureSerializer):
    feature_type = "arc-gis-adhoc"

    def get_id(self, obj):
        return obj.get("id", None)

    def get_geometry(self, obj):
        return obj["geometry"]

    def serialize_details(self, obj):
        return {
            **obj["feature"]["attributes"],
        }



from datetime import datetime
from typing import Any, Optional, Type

from dateutil.parser import parse
from django.contrib.gis.geos import (
    GEOSGeometry,
    MultiPoint,
    MultiPolygon,
    Point,
    Polygon,
)


def is_numeric(value):
    try:
        float(value)
        return True
    except ValueError:
        return False


def ensure_datetime(value: Any) -> Optional[datetime]:
    try:
        if not value:
            return None
        if isinstance(value, (int, float)) or (
            isinstance(value, str) and is_numeric(value)
        ):
            value = float(value)
            if value > 2**32:
                value /= 1000
            return datetime.fromtimestamp(value)
        return parse(value)
    except ValueError as ex:
        raise ValueError(f"Cannot convert {value} to datetime") from ex


def ensure_geometry(
    value: GEOSGeometry, expected_class: Type[GEOSGeometry]
) -> GEOSGeometry:
    if isinstance(value, expected_class):
        return value
    if issubclass(expected_class, MultiPolygon) and isinstance(value, Polygon):
        return MultiPolygon([value])
    if issubclass(expected_class, MultiPoint) and isinstance(value, Point):
        return MultiPoint([value])

    raise ValueError(
        f"Don't know how to convert {value.__class__.__name__} to {expected_class.__name__}"
    )



# Skytek Ltd - React
#
# Copyright (c) 2018 - 2022
# Author Skytek Ltd
#

from generic_map_api.routers import MapApiRouter

from . import views

router = MapApiRouter()
router.register("arc-gis", views.ParametrizedArcGisView, basename="arc-gis")

urlpatterns = router.urls



from jinja2.ext import Extension

from skytek_arcgis_integration.utils import is_numeric


def literal(value):
    if value is None or value == "None":
        return "None"

    if isinstance(value, (float, int)) or is_numeric(value):
        return str(value)

    if isinstance(value, bool) or value in ("True", "False"):
        return "True" if value else "False"

    if isinstance(value, dict):
        return (
            "{"
            + ", ".join(literal(k) + ": " + literal(v) for k, v in value.items())
            + ",}"
        )

    return '"' + str(value).replace('"', '\\"') + '"'


def kwargs(value):
    if not value:
        return ""
    assert isinstance(value, (dict))

    return ", ".join(str(k) + "=" + literal(v) for k, v in value.items()) + ","


class GeneratorExtension(Extension):
    def __init__(self, environment):
        super().__init__(environment)
        environment.filters["literal"] = literal
        environment.filters["kwargs"] = kwargs



from base64 import b64decode

from generic_map_api import params as api_params
from generic_map_api.views import MapFeaturesBaseView, ViewPort

from . import serializers
from .client import ArcGisClient


class ParametrizedArcGisView(MapFeaturesBaseView):
    display_name = "ArcGis"
    serializer = serializers.ParametrizedArcGisSerializer()

    query_params = {"layer_url": api_params.Text("Layer base url")}

    def get_items(self, viewport: ViewPort, params: dict):
        layer_url = params["layer_url"]
        arc_gis = ArcGisClient(layer_url)
        for item in arc_gis.get_feature_list(
            viewport.to_polygon() if viewport else None
        ):
            item["_layer_url"] = layer_url
            yield item

    def get_item(self, item_id):
        layer_url, feature_id = item_id.split(",")
        layer_url = b64decode(layer_url)
        arc_gis = ArcGisClient(layer_url)
        return arc_gis.get_feature(feature_id)


class AdHocBaseArcGisView(MapFeaturesBaseView):
    display_name = "ArcGis"
    serializer = serializers.AdHocBaseArcGisSerializer()

    layer_url = None

    def get_items(self, viewport: ViewPort, params: dict):
        layer_url = self.layer_url
        arc_gis = ArcGisClient(layer_url)
        return arc_gis.get_feature_list(viewport.to_polygon() if viewport else None)

    def get_item(self, item_id):
        layer_url = self.layer_url
        arc_gis = ArcGisClient(layer_url)
        arc_gis.format = "json"
        return arc_gis.get_feature(item_id)


def arcgis_view_factory(layer_url, feature_type=None, display_name=None):
    feature_type = feature_type or serializers.AdHocBaseArcGisSerializer.feature_type
    display_name = display_name or AdHocBaseArcGisView.display_name

    class AdHocSerializer(serializers.AdHocBaseArcGisSerializer):
        pass

    AdHocSerializer.feature_type = feature_type

    class AdHocView(AdHocBaseArcGisView):  # pylint: disable=too-many-ancestors
        pass

    AdHocView.layer_url = layer_url
    AdHocView.serializer = AdHocSerializer()
    AdHocView.display_name = display_name

    return AdHocView



import subprocess

from django.core.management.base import BaseCommand

from skytek_arcgis_integration.generator.generator import generate_django_module


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument("base-layer-url", type=str, nargs="?")
        parser.add_argument("--module-path", type=str)
        parser.add_argument("--model-name", type=str)
        parser.add_argument("--celery-app-path", type=str)
        parser.add_argument("--no-input", action="store_true")

    def handle(self, *args, **options):
        base_layer_url = options.get("base-layer-url")
        module_path = options.get("module_path")
        model_name = options.get("model_name")
        celery_app_path = options.get("celery_app_path")
        interactive = not options.get("no_input")

        output_dir, module_name = generate_django_module(
            base_layer_url=base_layer_url,
            module_path=module_path,
            model_name=model_name,
            celery_app_path=celery_app_path,
            interactive=interactive,
        )

        try:
            subprocess.run(
                ["python", "-m", "black", output_dir],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
                check=True,
            )
            self.stdout.write(
                self.style.SUCCESS(  # pylint: disable=no-member
                    "Successfully ran black"
                )
            )
        except subprocess.CalledProcessError:
            self.stdout.write(
                self.style.ERROR(  # pylint: disable=no-member
                    "Running black FAILED. No worries - you can run it manually later."
                )
            )

        try:
            subprocess.run(
                ["python", "-m", "isort", "--profile=black", output_dir],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
                check=True,
            )
            self.stdout.write(
                self.style.SUCCESS(  # pylint: disable=no-member
                    "Successfully ran isort"
                )
            )
        except subprocess.CalledProcessError:
            self.stdout.write(
                self.style.ERROR(  # pylint: disable=no-member
                    "Running isort FAILED.  No worries - you can run it manually later."
                )
            )

        self.stdout.write(
            f"All done. Now add '{module_name}' to INSTALLED_APPS in Django configuration, "
            f"run `makemigrations` and include '{module_name}.urls' in your urls configuration."
        )



import string
from dataclasses import asdict, dataclass
from os import path
from typing import List, Optional

from cookiecutter.main import cookiecutter
from django.conf import settings

from skytek_arcgis_integration.client import ArcGisClient


def ask_user(prompt, default=None):
    if default:
        message = f"{prompt} [{default}]: "
    else:
        message = f"{prompt}: "

    raw_value = input(message)
    if default and raw_value == "":
        raw_value = default
    return raw_value


def generate_django_module(  # pylint: disable=too-many-locals,too-many-branches
    base_layer_url=None,
    module_path=None,
    model_name=None,
    celery_app_path=None,
    interactive=True,
):
    given_base_layer_url = base_layer_url
    given_module_path = module_path
    given_model_name = model_name
    given_celery_app_path = celery_app_path

    if interactive and not given_base_layer_url:
        base_layer_url = ask_user("Enter base layer_url")

    client = ArcGisClient(base_layer_url)
    info = client.get_info()

    if info.get("type") != "Feature Layer":
        raise ValueError("Unsupported arcgis layer")

    template_data = TemplateData.from_arcgis_info(info)

    template_directory = path.join(path.dirname(__file__), "integration_template")
    main_django_app_module = ".".join(settings.SETTINGS_MODULE.split(".")[:-1])

    if given_module_path:
        module_name = given_module_path.split(".")[-1]
        full_module_name = given_module_path
    else:
        module_name = make_python_style_variable_name(info["name"])
        if module_name[-1] != "s":
            module_name += "s"
        full_module_name = f"arcgis.{module_name}"

        if interactive:
            full_module_name = ask_user("Enter full module path", full_module_name)

    if given_model_name:
        model_name = given_model_name
    else:
        model_name = make_python_style_class_name(info["name"])
        if model_name[-1].lower() == "s":
            model_name = model_name[:-1]

        if interactive:
            model_name = ask_user("Enter model name", model_name)

    if given_celery_app_path:
        celery_app = given_celery_app_path
    else:
        celery_app = f"{main_django_app_module}.celery.app"

        if interactive:
            celery_app = ask_user("Enter celery app path", celery_app)

    output_directory = path.join(settings.BASE_DIR, *full_module_name.split(".")[:-1])
    module_name = full_module_name.split(".")[-1]
    extra_context = {
        "module_name": module_name,
        "model_name": model_name,
        "celery_app": celery_app,
        "base_layer_url": base_layer_url,
        "specs": template_data.to_dict(),
    }

    cookiecutter(
        template=template_directory,
        output_dir=output_directory,
        extra_context=extra_context,
        no_input=True,
    )

    full_output_directory = path.join(settings.BASE_DIR, *full_module_name.split("."))

    return full_output_directory, full_module_name


TYPE_MAPPING = {
    "esriFieldTypeOID": "IntegerField",
    "esriFieldTypeSmallInteger": "IntegerField",
    "esriFieldTypeInteger": "IntegerField",
    "esriFieldTypeDate": "DateTimeField",
    "esriFieldTypeSingle": "FloatField",
    "esriFieldTypeDouble": "FloatField",
    "esriFieldTypeString": "CharField",
}


def make_python_style_variable_name(name):
    output = ""
    for previous_letter, letter in zip("_" + name, name):
        if letter == "_" and output[-1:] == "_":
            continue

        if (
            letter in string.ascii_uppercase
            and previous_letter in string.ascii_lowercase
        ):
            output += "_"

        output += letter.lower()
    return output


def make_python_style_class_name(name):
    output = ""
    for previous_letter, letter in zip("_" + name, name):
        if letter in string.ascii_letters + string.digits:
            if previous_letter not in string.ascii_letters:
                output += letter.upper()
            else:
                output += letter.lower()

    return output


@dataclass
class TemplateData:
    srid: int
    object_id_field: str
    object_id_field_in_model: str

    model_geometry_field: "Field"
    model_fields: List["Field"]

    def to_dict(self):
        return asdict(self)

    @classmethod
    def from_arcgis_info(cls, info_dict):
        model_geometry_field = cls.make_geometry_field(info_dict)
        model_fields = {
            field.api_field: field
            for field in [
                Field.from_esri_field(arcgis_field)
                for arcgis_field in info_dict["fields"]
            ]
            if field is not None
        }

        return cls(
            srid=info_dict["extent"]["spatialReference"]["wkid"],
            object_id_field=info_dict["objectIdField"],
            object_id_field_in_model=make_python_style_variable_name(
                info_dict["objectIdField"]
            ),
            model_geometry_field=model_geometry_field,
            model_fields=model_fields,
        )

    @classmethod
    def make_geometry_field(cls, info_dict):
        field_mapping = {
            "esriGeometryPoint": "PointField",
            "esriGeometryMultipoint": "MultiPointField",
            "esriGeometryPolygon": "MultiPolygonField",
        }
        model_field_type = field_mapping.get(info_dict["geometryType"], "GeometryField")
        return Field(
            api_field="",
            model_field_name="geometry",
            model_field_type=model_field_type,
            model_field_kwargs={"null": False},
        )


@dataclass
class Field:
    api_field: str
    model_field_name: str
    model_field_type: str
    model_field_kwargs: dict

    @classmethod
    def from_esri_field(cls, field_dict) -> Optional["Field"]:
        api_field = field_dict["name"]
        field_name = cls._transform_name(field_dict)
        field_type = cls._transform_type(field_dict)
        field_kwargs = cls._create_field_kwargs(field_dict)

        return cls(api_field, field_name, field_type, field_kwargs)

    @staticmethod
    def _transform_name(field_dict):
        name = field_dict["name"]
        name = make_python_style_variable_name(name)
        return name

    @staticmethod
    def _transform_type(field_dict):
        field_type = field_dict["type"]
        return TYPE_MAPPING.get(field_type, "CharField")

    @staticmethod
    def _create_field_kwargs(field_dict):
        kwargs = {}

        if "nullable" in field_dict and field_dict["nullable"]:
            kwargs["null"] = True

        if "alias" in field_dict and field_dict["alias"]:
            kwargs["verbose_name"] = field_dict["alias"]

        if field_dict["type"] == "esriFieldTypeString":
            kwargs["max_length"] = field_dict["length"]

        if field_dict["type"] not in TYPE_MAPPING:
            kwargs["max_length"] = 255

        return kwargs
